from decimal import Decimal

from sqlalchemy import Column, VARCHAR, INTEGER, BIGINT, ForeignKey, and_, event
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import composite, relation, foreign

from .._common import hybrid_property_gncnumeric, CallableList
from .._declbase import DeclarativeBaseGuid
from ..sa_extra import ChoiceType

TaxIncludedType = [
    (1, "YES"),
    (2, "NO"),
    (3, "USEGLOBAL")
]


class Address(object):
    """An Address object encapsulates information regarding an address in GnuCash.

    Attributes:
        name (str): self explanatory
        addr1 (str): self explanatory
        addr2 (str): self explanatory
        addr3 (str): self explanatory
        addr4 (str): self explanatory
        email (str): self explanatory
        fax (str): self explanatory
        phone (str): self explanatory
    """

    _address_fields = [
        "name",
        "addr1",
        "addr2",
        "addr3",
        "addr4",
        "email",
        "fax",
        "phone",
    ]

    def __init__(
        self,
        name="",
        addr1="",
        addr2="",
        addr3="",
        addr4="",
        email="",
        fax="",
        phone="",
    ):
        self.name = name
        self.addr1 = addr1
        self.addr2 = addr2
        self.addr3 = addr3
        self.addr4 = addr4
        self.email = email
        self.fax = fax
        self.phone = phone

    def __composite_values__(self):
        return (getattr(self, fld) for fld in Address._address_fields)

    def __eq__(self, other):
        return isinstance(other, Address) and all(
            getattr(other, fld) == getattr(self, fld) for fld in Address._address_fields
        )

    def __ne__(self, other):
        return not self.__eq__(other)


class Person:
    """A mixin declaring common field for Customer, Vendor and Employee"""

    active = Column("active", INTEGER(), nullable=False)

    id = Column("id", VARCHAR(length=2048), nullable=False)

    addr_name = Column("addr_name", VARCHAR(length=1024))
    addr_addr1 = Column("addr_addr1", VARCHAR(length=1024))
    addr_addr2 = Column("addr_addr2", VARCHAR(length=1024))
    addr_addr3 = Column("addr_addr3", VARCHAR(length=1024))
    addr_addr4 = Column("addr_addr4", VARCHAR(length=1024))
    addr_phone = Column("addr_phone", VARCHAR(length=128))
    addr_fax = Column("addr_fax", VARCHAR(length=128))
    addr_email = Column("addr_email", VARCHAR(length=256))

    @declared_attr
    def address(cls):
        return composite(
            Address,
            cls.addr_name,
            cls.addr_addr1,
            cls.addr_addr2,
            cls.addr_addr3,
            cls.addr_addr4,
            cls.addr_email,
            cls.addr_fax,
            cls.addr_phone,
        )

    @declared_attr
    def currency_guid(cls):
        return Column(
            "currency",
            VARCHAR(length=32),
            ForeignKey("commodities.guid"),
            nullable=False,
        )

    @declared_attr
    def currency(cls):
        return relation("Commodity")

    # hold the name of the counter to use for id
    _counter_name = None

    def _assign_id(self):
        if not self.id:
            cnt = getattr(self.book, self._counter_name) + 1
            setattr(self.book, self._counter_name, cnt)
            self.id = "{:06d}".format(cnt)

    def object_to_validate(self, change):
        yield self

    def validate(self):
        if self.id is None:
            raise ValueError(
                "You cannot flush a {} that is not added to a book".format(
                    self.__class__.__name__
                )
            )

    def on_book_add(self):
        self._assign_id()
        for job in self.jobs:
            job.on_book_add()

    def __str__(self):
        return "{}<{}:{}>".format(self.__class__.__name__, self.id, self.name)

    @classmethod
    def __declare_last__(cls):
        from .invoice import Job

        owner_type = PersonType.get(cls, None)
        if owner_type and not hasattr(cls, "jobs"):
            cls.jobs = relation('Job',
                primaryjoin=and_(
                    cls.guid == foreign(Job.owner_guid),
                    owner_type == Job.owner_type,
                ),
                cascade="all, delete-orphan",
                collection_class=CallableList,
            )

            @event.listens_for(cls.jobs, "append")
            def add(target, value, initiator):
                value.owner_type = owner_type
                value.owner_guid = target.guid
                value._assign_id()


class Customer(Person, DeclarativeBaseGuid):
    """
    A GnuCash Customer

    Attributes:
        name (str): name of the Customer
        id (str): autonumber id with 5 digits (initialised to book.counter_customer + 1)
        notes (str): notes
        active (int): 1 if the customer is active, 0 otherwise
        discount (:class:`decimal.Decimal`): see Gnucash documentation
        credit (:class:`decimal.Decimal`): see Gnucash documentation
        currency (:class:`piecash.core.commodity.Commodity`): the currency of the customer
        tax_override (int): 1 if tax override, 0 otherwise
        address (:class:`Address`): the address of the customer
        shipping_address (:class:`Address`): the shipping address of the customer
        tax_included (str): 'yes', 'no', 'use global'
        taxtable (:class:`piecash.business.tax.TaxTable`): tax table of the customer
        term (:class:`piecash.business.invoice.Billterm`): bill term of the customer
    """

    __tablename__ = "customers"

    __table_args__ = {}

    # column definitions
    name = Column("name", VARCHAR(length=2048), nullable=False)
    # id is nullable as it is set during validation (happening after flush)
    notes = Column("notes", VARCHAR(length=2048), nullable=False)
    _discount_num = Column("discount_num", BIGINT(), nullable=False)
    _discount_denom = Column("discount_denom", BIGINT(), nullable=False)
    discount = hybrid_property_gncnumeric(_discount_num, _discount_denom)
    _credit_num = Column("credit_num", BIGINT(), nullable=False)
    _credit_denom = Column("credit_denom", BIGINT(), nullable=False)
    credit = hybrid_property_gncnumeric(_credit_num, _credit_denom)
    tax_override = Column("tax_override", INTEGER(), nullable=False)

    shipaddr_name = Column("shipaddr_name", VARCHAR(length=1024))
    shipaddr_addr1 = Column("shipaddr_addr1", VARCHAR(length=1024))
    shipaddr_addr2 = Column("shipaddr_addr2", VARCHAR(length=1024))
    shipaddr_addr3 = Column("shipaddr_addr3", VARCHAR(length=1024))
    shipaddr_addr4 = Column("shipaddr_addr4", VARCHAR(length=1024))
    shipaddr_phone = Column("shipaddr_phone", VARCHAR(length=128))
    shipaddr_fax = Column("shipaddr_fax", VARCHAR(length=128))
    shipaddr_email = Column("shipaddr_email", VARCHAR(length=256))
    shipping_address = composite(
        Address,
        shipaddr_name,
        shipaddr_addr1,
        shipaddr_addr2,
        shipaddr_addr3,
        shipaddr_addr4,
        shipaddr_email,
        shipaddr_fax,
        shipaddr_phone,
    )

    term_guid = Column("terms", VARCHAR(length=32), ForeignKey("billterms.guid"))
    tax_included = Column("tax_included", ChoiceType(TaxIncludedType))
    taxtable_guid = Column("taxtable", VARCHAR(length=32), ForeignKey("taxtables.guid"))

    # relation definitions
    taxtable = relation("Taxtable")
    term = relation("Billterm")

    def __init__(
        self,
        name,
        currency,
        id=None,
        notes="",
        active=1,
        tax_override=0,
        credit=Decimal(0),
        discount=Decimal(0),
        taxtable=None,
        address=None,
        shipping_address=None,
        tax_included="USEGLOBAL",
        book=None,
    ):
        self.name = name
        self.currency = currency
        self.notes = notes
        self.active = active
        self.credit = credit
        self.discount = discount
        self.tax_included = tax_included
        self.taxtable = taxtable
        self.tax_override = tax_override
        if address is None:
            address = Address(name=name)
        self.address = address
        if shipping_address is None:
            shipping_address = Address("")
        self.shipping_address = shipping_address

        if book and id is None:
            book.add(self)
        elif id is not None:
            if isinstance(id, int):
                self.id = str(id)
            else:
                self.id = id

    _counter_name = "counter_customer"


class Employee(Person, DeclarativeBaseGuid):
    """
    A GnuCash Employee

    Attributes:
        name (str): name of the Employee
        id (str): autonumber id with 5 digits (initialised to book.counter_employee + 1)
        language (str): language
        active (int): 1 if the employee is active, 0 otherwise
        workday (:class:`decimal.Decimal`): see Gnucash documentation
        rate (:class:`decimal.Decimal`): see Gnucash documentation
        currency (:class:`piecash.core.commodity.Commodity`): the currency of the employee
        address (:class:`Address`): the address of the employee
        creditcard_account (:class:`piecash.core.account.Account`): credit card account for the employee

    """

    __tablename__ = "employees"

    __table_args__ = {}

    # column definitions
    name = Column("username", VARCHAR(length=2048), nullable=False)
    # id is nullable as it is set during validation (happening after flush)
    language = Column("language", VARCHAR(length=2048), nullable=False)
    acl = Column("acl", VARCHAR(length=2048), nullable=False)
    ccard_guid = Column("ccard_guid", VARCHAR(length=32), ForeignKey("accounts.guid"))
    _workday_num = Column("workday_num", BIGINT(), nullable=False)
    _workday_denom = Column("workday_denom", BIGINT(), nullable=False)
    workday = hybrid_property_gncnumeric(_workday_num, _workday_denom)
    _rate_num = Column("rate_num", BIGINT(), nullable=False)
    _rate_denom = Column("rate_denom", BIGINT(), nullable=False)
    rate = hybrid_property_gncnumeric(_rate_num, _rate_denom)

    # relation definitions
    creditcard_account = relation("Account")

    def __init__(
        self,
        name,
        currency,
        creditcard_account=None,
        id=None,
        active=1,
        acl="",
        language="",
        workday=Decimal(0),
        rate=Decimal(0),
        address=None,
        book=None,
    ):
        self.name = name
        self.currency = currency
        self.active = active
        self.workday = workday
        self.rate = rate
        self.acl = acl
        self.language = language
        self.creditcard_account = creditcard_account
        if address is None:
            address = Address(name=name)
        self.address = address

        if book and id is None:
            book.add(self)

        elif id is not None:
            if isinstance(id, int):
                self.id = str(id)
            else:
                self.id = id

    _counter_name = "counter_employee"

    def on_book_add(self):
        self._assign_id()


class Vendor(Person, DeclarativeBaseGuid):
    """
    A GnuCash Vendor

    Attributes:
        name (str): name of the Vendor
        id (str): autonumber id with 5 digits (initialised to book.counter_vendor + 1)
        notes (str): notes
        active (int): 1 if the vendor is active, 0 otherwise
        currency (:class:`piecash.core.commodity.Commodity`): the currency of the vendor
        tax_override (int): 1 if tax override, 0 otherwise
        address (:class:`Address`): the address of the vendor
        tax_included (str): 'YES', 'NO', 'USEGLOBAL'
        taxtable (:class:`piecash.business.tax.TaxTable`): tax table of the vendor
        term (:class:`piecash.business.invoice.Billterm`): bill term of the vendor
    """

    __tablename__ = "vendors"

    __table_args__ = {}

    # column definitions
    name = Column("name", VARCHAR(length=2048), nullable=False)
    # id is nullable as it is set during validation (happening after flush)
    notes = Column("notes", VARCHAR(length=2048), nullable=False)
    tax_override = Column("tax_override", INTEGER(), nullable=False)

    term_guid = Column("terms", VARCHAR(length=32), ForeignKey("billterms.guid"))
    tax_included = Column("tax_inc", VARCHAR(length=2048))
    tax_table_guid = Column(
        "tax_table", VARCHAR(length=32), ForeignKey("taxtables.guid")
    )

    # relation definitions
    taxtable = relation("Taxtable")
    term = relation("Billterm")

    def __init__(
        self,
        name,
        currency,
        id=None,
        notes="",
        active=1,
        tax_override=0,
        taxtable=None,
        credit=Decimal(0),
        discount=Decimal(0),
        address=None,
        tax_included="USEGLOBAL",
        book=None,
    ):
        self.name = name
        self.currency = currency
        self.notes = notes
        self.active = active
        self.credit = credit
        self.discount = discount
        self.tax_included = tax_included
        self.taxtable = taxtable
        self.tax_override = tax_override
        if address is None:
            address = Address(name=name)
        self.address = address

        if book and id is None:
            book.add(self)
        elif id is not None:
            if isinstance(id, int):
                self.id = str(id)
            else:
                self.id = id

    _counter_name = "counter_vendor"


PersonType = {Vendor: 4, Customer: 2}



#!/usr/bin/env python
##  @file
#   @brief Example Script simple sqlite create
#   @ingroup python_bindings_examples

from __future__ import print_function
import os

from piecash import create_book, Account, Commodity, open_book
from piecash.core.factories import create_currency_from_ISO

filename = os.path.abspath("test.blob")
if os.path.exists(filename):
    os.remove(filename)

with create_book(filename) as book:
    a = Account(
        parent=book.root_account,
        name="wow",
        type="ASSET",
        commodity=create_currency_from_ISO("CAD"),
    )

    book.save()

with open_book(filename) as book:
    print(book.root_account.children)
    print(book.commodities.get(mnemonic="CAD"))

os.remove(filename)



#!/usr/bin/env python

import click

from piecash.scripts.cli import cli


@cli.command()
@click.argument("book", type=click.Path(exists=True))
@click.argument("entities", type=click.Choice(["customers", "vendors", "prices"]))
@click.option(
    "--output",
    type=click.File("w"),
    default="-",
    help="File to which to export the data (default=stdout)",
)
@click.option(
    "--inactive",
    is_flag=True,
    default=False,
    help="Include inactive entities (for vendors and customers)",
)
def export(book, entities, output, inactive):
    """Exports GnuCash ENTITIES.

    This scripts export ENTITIES from the BOOK in a CSV format.
    When possible, it exports in a format that can be used to import the data into GnuCash.

    \b
    Remarks:
    - for customers and vendors, the format does not include an header
    - for prices, the format can be used with the `piecash import` command.
    """
    from piecash import open_book

    with open_book(book, open_if_lock=True) as book:
        if entities == "prices":
            output.write(
                "date,type,value,value_num, value_denom, currency,commodity,source\n"
            )
            output.writelines(
                "{p.date:%Y-%m-%d},{p.type},{p.value},{p._value_num},{p._value_denom},{p.currency.mnemonic},{p.commodity.mnemonic},{p.source}\n".format(
                    p=p
                )
                for p in book.prices
            )

        elif entities in ["customers", "vendors"]:
            columns = (
                "id, name, addr_name, addr_addr1, addr_addr2, addr_addr3, addr_addr4, "
                "addr_phone, addr_fax, addr_email, notes, shipaddr_name, "
                "shipaddr_addr1, shipaddr_addr2, shipaddr_addr3, shipaddr_addr4, "
                "shipaddr_phone, shipaddr_fax, shipaddr_email".split(", ")
            )
            separator = ";"

            filter_entity = (lambda e: True) if inactive else (lambda e: e.active)

            # open the book
            res = "\n".join(
                [
                    separator.join(getattr(v, fld, "") for fld in columns)
                    for v in getattr(book, entities)
                    if filter_entity(v)
                ]
            )

            output.write(res)

if __name__ == '__main__':
    cli()


import uuid

from sqlalchemy import Column, VARCHAR, event
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import relation, foreign, object_session

from ._common import CallableList
from .kvp import DictWrapper, Slot
from .sa_extra import DeclarativeBase


class DeclarativeBaseGuid(DictWrapper, DeclarativeBase):
    __abstract__ = True

    #: the unique identifier of the object
    guid = Column(
        "guid",
        VARCHAR(length=32),
        primary_key=True,
        nullable=False,
        default=lambda: uuid.uuid4().hex,
    )

    @declared_attr
    def slots(cls):
        rel = relation(
            "Slot",
            primaryjoin=foreign(Slot.obj_guid) == cls.guid,
            cascade="all, delete-orphan",
            collection_class=CallableList,
        )

        return rel



import uuid
from sqlalchemy import Column, VARCHAR, BIGINT, INTEGER, ForeignKey
from sqlalchemy.orm import relation

from .._common import hybrid_property_gncnumeric, CallableList
from .._declbase import DeclarativeBaseGuid, DeclarativeBase
from ..sa_extra import ChoiceType


class Taxtable(DeclarativeBaseGuid):
    __tablename__ = "taxtables"

    __table_args__ = {}

    # column definitions
    guid = Column(
        "guid",
        VARCHAR(length=32),
        primary_key=True,
        nullable=False,
        default=lambda: uuid.uuid4().hex,
    )
    name = Column("name", VARCHAR(length=50), nullable=False)
    refcount = Column("refcount", BIGINT(), nullable=False)
    invisible = Column("invisible", INTEGER(), nullable=False)
    parent_guid = Column("parent", VARCHAR(length=32), ForeignKey("taxtables.guid"))

    # relation definitions
    entries = relation(
        "TaxtableEntry",
        back_populates="taxtable",
        cascade="all, delete-orphan",
        collection_class=CallableList,
    )
    children = relation(
        "Taxtable",
        back_populates="parent",
        cascade="all, delete-orphan",
        collection_class=CallableList,
    )
    parent = relation(
        "Taxtable",
        back_populates="children",
        remote_side=guid,
    )

    def __init__(self, name, entries=None):
        self.name = name
        self.refcount = 0
        self.invisible = 0
        if entries is not None:
            self.entries[:] = entries

    def __str__(self):
        if self.entries:
            return "TaxTable<{}:{}>".format(
                self.name, [te.__str__() for te in self.entries]
            )
        else:
            return "TaxTable<{}>".format(self.name)


class TaxtableEntry(DeclarativeBase):
    __tablename__ = "taxtable_entries"

    __table_args__ = {"sqlite_autoincrement": True}

    # column definitions
    id = Column("id", INTEGER(), primary_key=True, nullable=False, autoincrement=True)
    taxtable_guid = Column(
        "taxtable", VARCHAR(length=32), ForeignKey("taxtables.guid"), nullable=False
    )
    account_guid = Column(
        "account", VARCHAR(length=32), ForeignKey("accounts.guid"), nullable=False
    )
    _amount_num = Column("amount_num", BIGINT(), nullable=False)
    _amount_denom = Column("amount_denom", BIGINT(), nullable=False)
    amount = hybrid_property_gncnumeric(_amount_num, _amount_denom)
    type = Column("type", ChoiceType({1: "value", 2: "percentage"}), nullable=False)

    # relation definitions
    taxtable = relation("Taxtable", back_populates="entries")
    account = relation("Account")

    def __init__(self, type, amount, account, taxtable=None):
        self.type = type
        self.amount = amount
        self.account = account
        if taxtable:
            self.taxtable = taxtable

    def __str__(self):
        return "TaxEntry<{} {} in {}>".format(self.amount, self.type, self.account.name)



# -*- coding: utf-8 -*-
"""Project metadata

Information describing the project.
"""

# The package name, which is also the "UNIX name" for the project.
package = "piecash"
project = "Python interface to GnuCash documents"
project_no_spaces = project.replace(" ", "")
version = "1.2.1"
description = "A pythonic interface to GnuCash SQL documents."
authors = ["sdementen"]
authors_string = ", ".join(authors)
emails = ["sdementen@gmail.com"]
license = "MIT"
copyright = "2014 " + authors_string
url = "https://github.com/sdementen/piecash"



# coding=utf-8
from __future__ import unicode_literals

from collections import defaultdict
from datetime import datetime, date, time
from decimal import Decimal

import pytest

from piecash import Transaction, Split, GncImbalanceError, GncValidationError, Lot
from test_helper import (
    db_sqlite_uri,
    db_sqlite,
    new_book,
    new_book_USD,
    book_uri,
    book_basic,
    book_transactions,
)

# dummy line to avoid removing unused symbols

a = (
    db_sqlite_uri,
    db_sqlite,
    new_book,
    new_book_USD,
    book_uri,
    book_basic,
    book_transactions,
)


class TestTransaction_create_transaction(object):
    def test_create_basictransaction_neutraltime(self, book_basic):
        EUR = book_basic.commodities(namespace="CURRENCY")
        racc = book_basic.root_account
        a = book_basic.accounts(name="asset")
        e = book_basic.accounts(name="exp")

        tr = Transaction(
            currency=EUR,
            description="wire from Hélène",
            notes="on St-Eugène day",
            post_date=date(2014, 1, 1),
            splits=[
                Split(account=a, value=100, memo="mémo asset"),
                Split(account=e, value=-100, memo="mémo exp"),
            ],
        )

        assert isinstance(tr.post_date, date)

        book_basic.flush()
        book_basic.validate()

        assert isinstance(tr.post_date, date)

    def test_create_basictransaction_validation_date(self, book_basic):
        EUR = book_basic.commodities(namespace="CURRENCY")
        racc = book_basic.root_account
        a = book_basic.accounts(name="asset")
        e = book_basic.accounts(name="exp")

        splits = [
            Split(account=a, value=100, memo="mémo asset"),
            Split(account=e, value=-10, memo="mémo exp"),
        ]

        with pytest.raises(GncValidationError):
            tr = Transaction(
                currency=EUR,
                description="wire from Hélène",
                notes="on St-Eugène day",
                post_date=datetime(2014, 1, 1),
                enter_date=datetime(2014, 1, 1),
                splits=splits,
            )

        with pytest.raises(GncValidationError):
            tr = Transaction(
                currency=EUR,
                description="wire from Hélène",
                notes="on St-Eugène day",
                post_date=datetime(2014, 1, 1),
                enter_date=time(10, 59, 00),
                splits=splits,
            )

        with pytest.raises(GncValidationError):
            tr = Transaction(
                currency=EUR,
                description="wire from Hélène",
                notes="on St-Eugène day",
                post_date=date(2014, 1, 1),
                enter_date=date(2014, 1, 1),
                splits=splits,
            )

        tr = Transaction(
            currency=EUR,
            description="wire from Hélène",
            notes="on St-Eugène day",
            post_date=None,
            enter_date=None,
            splits=splits,
        )

        with pytest.raises(GncImbalanceError):
            book_basic.flush()
            book_basic.validate()

    def test_create_basictransaction(self, book_basic):
        EUR = book_basic.commodities(namespace="CURRENCY")
        racc = book_basic.root_account
        a = book_basic.accounts(name="asset")
        e = book_basic.accounts(name="exp")

        tr = Transaction(
            currency=EUR,
            description="wire from Hélène",
            notes="on St-Eugène day",
            post_date=date(2014, 1, 1),
            enter_date=datetime(2014, 1, 1),
            splits=[
                Split(account=a, value=100, memo="mémo asset"),
                Split(account=e, value=-10, memo="mémo exp"),
            ],
        )
        # check issue with balance
        with pytest.raises(GncImbalanceError):
            book_basic.flush()
            book_basic.validate()

        # adjust balance
        Split(account=e, value=-90, memo="missing exp", transaction=tr)
        book_basic.flush()

        # check no issue with str
        assert str(tr)
        assert str(tr.splits)
        assert repr(tr)
        assert repr(tr.splits)
        assert tr.notes == "on St-Eugène day"

    def test_create_basictransaction_splitfirst(self, book_basic):
        EUR = book_basic.commodities(namespace="CURRENCY")
        racc = book_basic.root_account
        a = book_basic.accounts(name="asset")
        e = book_basic.accounts(name="exp")
        s = Split(account=a, value=Decimal(1))
        assert repr(s)

    def test_create_cdtytransaction(self, book_basic):
        EUR = book_basic.commodities(namespace="CURRENCY")
        racc = book_basic.root_account
        a = book_basic.accounts(name="asset")
        s = book_basic.accounts(name="broker")

        tr = Transaction(
            currency=EUR,
            description="buy stock",
            notes="on St-Eugène day",
            post_date=date(2014, 1, 2),
            enter_date=datetime(2014, 1, 3),
            splits=[
                Split(account=a, value=100, memo="mémo asset"),
                Split(account=s, value=-90, memo="mémo brok"),
            ],
        )

        # check issue with quantity for broker split not defined
        with pytest.raises(GncValidationError):
            book_basic.validate()

        sb = tr.splits(account=s)
        sb.quantity = 15

        # check issue with quantity not same sign as value
        with pytest.raises(GncValidationError):
            book_basic.validate()

        sb.quantity = -15

        # verify imbalance issue
        with pytest.raises(GncImbalanceError):
            book_basic.validate()

        # adjust balance
        Split(account=a, value=-10, memo="missing asset corr", transaction=tr)
        book_basic.save()
        assert str(sb)
        assert str(sb)

        # changing currency of an existing transaction is not allowed
        tr.currency = book_basic.currencies(mnemonic="USD")
        with pytest.raises(GncValidationError):
            book_basic.validate()
        book_basic.cancel()

        # check sum of quantities are not balanced per commodity but values are
        d = defaultdict(lambda: Decimal(0))
        for sp in tr.splits:
            assert sp.quantity == sp.value or sp.account != a
            d[sp.account.commodity] += sp.quantity
            d["cur"] += sp.value
        assert d["cur"] == 0
        assert all([v != 0 for k, v in d.items() if k != "cur"])

    def test_create_split_overflow(self, book_basic):
        a = book_basic.accounts(name="asset")

        # raise error as Transaction has a non CURRENCY commodity
        with pytest.raises(TypeError):
            sp = Split(account=a, value=1.0 / 3.0, quantity=10, memo="mémo asset")

        with pytest.raises(ValueError):
            sp = Split(
                account=a, value=Decimal(1) / Decimal(3), quantity=10, memo="mémo asset"
            )

        sp = Split(
            account=a,
            value=Decimal(1234567890123455678),
            quantity=10,
            memo="mémo asset",
        )

        with pytest.raises(ValueError):
            sp = Split(
                account=a,
                value=Decimal(1234567890123455678901234),
                quantity=10,
                memo="mémo asset",
            )

    def test_create_cdtytransaction_cdtycurrency(self, book_basic):
        EUR = book_basic.commodities(namespace="CURRENCY")
        racc = book_basic.root_account
        a = book_basic.accounts(name="asset")
        s = book_basic.accounts(name="broker")

        tr = Transaction(
            currency=s.commodity,
            description="buy stock",
            notes="on St-Eugène day",
            post_date=date(2014, 1, 2),
            enter_date=datetime(2014, 1, 3),
            splits=[
                Split(account=a, value=100, quantity=10, memo="mémo asset"),
                Split(account=s, value=-100, quantity=-10, memo="mémo brok"),
            ],
        )
        # raise error as Transaction has a non CURRENCY commodity
        with pytest.raises(GncValidationError):
            book_basic.validate()

    def test_create_cdtytransaction_tradingaccount(self, book_basic):
        EUR = book_basic.commodities(namespace="CURRENCY")
        racc = book_basic.root_account
        a = book_basic.accounts(name="asset")
        s = book_basic.accounts(name="broker")

        book_basic.use_trading_accounts = True
        tr = Transaction(
            currency=EUR,
            description="buy stock",
            notes="on St-Eugène day",
            post_date=date(2014, 1, 2),
            enter_date=datetime(2014, 1, 3),
            splits=[
                Split(account=a, value=100, memo="mémo asset"),
                Split(account=s, value=-100, quantity=-15, memo="mémo brok"),
            ],
        )
        book_basic.validate()

        assert "{}".format(tr) == "Transaction<[EUR] 'buy stock' on 2014-01-02>"
        assert "{}".format(s) == "Account<asset:broker[ïoà]>"
        assert (
            "{}".format(tr.splits(account=s))
            == "Split<Account<asset:broker[ïoà]> -100 EUR [-15 ïoà]>"
        )
        assert "{}".format(tr.splits(account=a)) == "Split<Account<asset[EUR]> 100 EUR>"

        # check sum of quantities are all balanced per commodity as values are
        d = defaultdict(lambda: Decimal(0))
        for sp in tr.splits:
            assert sp.quantity == sp.value or sp.account != a
            d[sp.account.commodity] += sp.quantity
            d["cur"] += sp.value

        assert d["cur"] == 0
        assert all([v == Decimal(0) for k, v in d.items() if k != "cur"])

        # change existing quantity
        sp = tr.splits(memo="mémo brok")
        sp.quantity += 1
        book_basic.validate()

        # check sum of quantities are all balanced per commodity as values are
        d = defaultdict(lambda: Decimal(0))
        for sp in tr.splits:
            assert sp.quantity == sp.value or sp.account != a
            d[sp.account.commodity] += sp.quantity
            d["cur"] += sp.value
        assert d["cur"] == 0
        assert all([v == Decimal(0) for k, v in d.items() if k != "cur"])

    def test_tag_split_zero_quantity(self, book_transactions):
        broker = book_transactions.accounts(name="broker")
        asset = book_transactions.accounts.get(name="asset")
        inc = book_transactions.accounts.get(name="inc")
        currency = book_transactions.default_currency

        value = Decimal(250)
        splits = [
            Split(asset, value),
            Split(inc, -value),
            Split(
                broker, value=0, quantity=0
            ),  # tag split for assigning dividend income to stock
        ]

        Transaction(currency, description="Dividend income", splits=splits)

        book_transactions.validate()

    def test_tag_split_zero_quantity_with_value(self, book_transactions):
        broker = book_transactions.accounts(name="broker")
        inc = book_transactions.accounts.get(name="inc")
        value = Decimal(250)

        # Transaction recording capital gains.
        splits = [
            Split(broker, value, quantity=0),
            Split(inc, -value),
        ]
        Transaction(inc.commodity, description="Capital gains", splits=splits)
        book_transactions.validate()

        # Transaction recording capital loss.
        splits = [
            Split(broker, -value, quantity=0),
            Split(inc, value),
        ]
        Transaction(inc.commodity, description="Capital loss", splits=splits)
        book_transactions.validate()

        # Do the same tests with a -0.0 quantity. This Decimal has is_signed=True.
        mzero = Decimal("-0.00")

        # Transaction recording capital gains.
        splits = [
            Split(broker, value, quantity=mzero),
            Split(inc, -value),
        ]
        Transaction(inc.commodity, description="Capital gains", splits=splits)
        book_transactions.validate()

        # Transaction recording capital loss.
        splits = [
            Split(broker, -value, quantity=mzero),
            Split(inc, value),
        ]
        Transaction(inc.commodity, description="Capital loss", splits=splits)
        book_transactions.validate()

    def test_tag_split_zero_value(self, book_transactions):
        broker = book_transactions.accounts(name="broker")
        asset = book_transactions.accounts.get(name="asset")
        currency = book_transactions.default_currency

        # Give away 250 shares for free.
        quantity = Decimal(-250)
        splits = [
            Split(asset, 0),
            Split(broker, 0, quantity=quantity),
        ]

        Transaction(currency, description="donation", splits=splits)

        book_transactions.validate()


class TestTransaction_lots(object):
    def test_create_simpletlot_addsplits(self, book_basic):
        EUR = book_basic.commodities(namespace="CURRENCY")
        racc = book_basic.root_account
        a = book_basic.accounts(name="asset")
        s = book_basic.accounts(name="broker")
        l = Lot(title="test mé", account=s, notes="ïlya")
        for i, am in enumerate([45, -35, -20]):
            tr = Transaction(
                currency=EUR,
                description="trade stock",
                notes="àçö",
                post_date=date(2014, 1, 1 + i),
                enter_date=datetime(2014, 1, 1 + i),
                splits=[
                    Split(account=a, value=am * 10, memo="mémo asset"),
                    Split(
                        account=s, value=-am * 10, quantity=-am, memo="mémo brok", lot=l
                    ),
                ],
            )
        book_basic.flush()

    def test_create_simpletlot_initialsplits(self, book_basic):
        EUR = book_basic.commodities(namespace="CURRENCY")
        racc = book_basic.root_account
        a = book_basic.accounts(name="asset")
        s = book_basic.accounts(name="broker")
        sp = []
        for i, am in enumerate([45, -35, -20]):
            tr = Transaction(
                currency=EUR,
                description="trade stock",
                notes="àçö",
                post_date=date(2014, 1, 1 + i),
                enter_date=datetime(2014, 1, 1 + i),
                splits=[
                    Split(account=a, value=am * 10, memo="mémo asset"),
                    Split(account=s, value=-am * 10, quantity=-am, memo="mémo brok"),
                ],
            )
            sp.append(tr.splits(account=s))

        l = Lot(title="test mé", account=s, notes="ïlya", splits=sp)
        book_basic.flush()

    def test_create_closedlot_addsplits(self, book_basic):
        EUR = book_basic.commodities(namespace="CURRENCY")
        racc = book_basic.root_account
        a = book_basic.accounts(name="asset")
        s = book_basic.accounts(name="broker")
        l = Lot(title="test mé", account=s, notes="ïlya")
        l.is_closed = 1
        # raise valueerror as lot is closed
        with pytest.raises(ValueError):
            tr = Transaction(
                currency=EUR,
                description="trade stock",
                notes="àçö",
                post_date=date(2014, 1, 1),
                enter_date=datetime(2014, 1, 1),
                splits=[
                    Split(account=a, value=10, memo="mémo asset"),
                    Split(account=s, value=-10, quantity=-2, memo="mémo brok", lot=l),
                ],
            )

    def test_create_simplelot_inconsistentaccounts(self, book_basic):
        EUR = book_basic.commodities(namespace="CURRENCY")
        racc = book_basic.root_account
        a = book_basic.accounts(name="asset")
        s = book_basic.accounts(name="broker")
        l = Lot(title="test mé", account=a, notes="ïlya")
        # raise valueerror as split account not the same as lot account
        tr = Transaction(
            currency=EUR,
            description="trade stock",
            notes="àçö",
            post_date=date(2014, 1, 1),
            enter_date=datetime(2014, 1, 1),
            splits=[
                Split(account=a, value=10, memo="mémo asset"),
                Split(account=s, value=-10, quantity=-2, memo="mémo brok", lot=l),
            ],
        )

        with pytest.raises(ValueError):
            book_basic.validate()


class TestTransaction_changes(object):
    def test_delete__replace_existing_split(self, book_transactions):
        s = len(book_transactions.splits)
        transaction = book_transactions.transactions(description="my revenue")
        split = transaction.splits(value=1000)
        assert len(transaction.splits) == 2
        splits = [
            Split(
                account=split.account,
                value=split.value,
                quantity=split.quantity,
            )
            for split in list(transaction.splits)
        ]

        assert len(transaction.splits) == 2
        transaction.splits[:] = splits
        assert split.transaction is None
        assert len(transaction.splits) == 2
        book_transactions.flush()
        book_transactions.save()
        ns = len(book_transactions.splits)
        assert ns == s

    def test_delete__replace_existing_split__split_with_transaction(
        self, book_transactions
    ):
        s = len(book_transactions.splits)
        transaction = book_transactions.transactions(description="my revenue")
        split = transaction.splits(value=1000)
        split_guid = split.guid
        assert len(transaction.splits) == 2
        splits = list(transaction.splits)
        del transaction.splits[:]

        splits = [
            Split(
                account=split.account,
                value=split.value * 2,
                quantity=split.quantity * 2,
                transaction=transaction,
            )
            for split in splits
        ]

        assert len(transaction.splits) == 2
        transaction.splits[:] = splits
        assert split.transaction is None
        assert len(transaction.splits) == 2
        book_transactions.flush()
        book_transactions.save()
        # check that previous split has been deleted
        with pytest.raises(KeyError, match="Could not find object with {'guid'"):
            split = book_transactions.splits(guid=split_guid)
        ns = len(book_transactions.splits)
        assert ns == s

    def test_delete_existing_transaction(self, book_transactions):
        l = len(book_transactions.transactions)
        s = len(book_transactions.splits)
        tr = book_transactions.transactions(description="my revenue")
        book_transactions.delete(tr)
        book_transactions.save()
        nl = len(book_transactions.transactions)
        ns = len(book_transactions.splits)
        assert nl == l - 1
        assert ns == s - 2

    def test_change_cdty_split_price(self, book_transactions):
        tr = book_transactions.transactions(description="my purchase of stock")
        sp = tr.splits(account=book_transactions.accounts(name="broker"))

        assert len(book_transactions.prices) == 6

        p = [p for p in book_transactions.prices if p.date.day == 29][0]
        p_expected = (sp.value / sp.quantity).quantize(Decimal("0.000001"))
        assert p.value == p_expected

        # changing the quantity of the split should NOT change the existing price
        sp.quantity = (5, 1)

        book_transactions.validate()

        p_not_expected = (sp.value / sp.quantity).quantize(Decimal("0.000001"))

        assert len(book_transactions.prices) == 6
        assert p.value == p_expected
        assert p_expected != p_not_expected

        # changing the post date of the transaction of the split should create a new price
        tr.post_date = date(2015, 1, 29)
        book_transactions.validate()
        book_transactions.flush()
        assert len(book_transactions.prices) == 7


class TestSplit_credit_debit(object):
    def test_delete_existing_transaction(self, book_transactions):
        asset = book_transactions.accounts.get(name="asset")
        inc = book_transactions.accounts.get(name="inc")
        exp = book_transactions.accounts.get(name="exp")

        # revenue item
        revenue = book_transactions.transactions(description="my revenue")
        revenue_split_asset = revenue.splits(account=asset)
        revenue_split_inc = revenue.splits(account=inc)
        assert revenue_split_asset.is_credit is False
        assert revenue_split_asset.is_debit is True
        assert revenue_split_inc.is_credit is True
        assert revenue_split_inc.is_debit is False

        # expense item
        expense = book_transactions.transactions(description="my expense")
        expense_split_asset = expense.splits(account=asset)
        expense_split_inc = expense.splits(account=exp)
        assert expense_split_asset.is_credit is True
        assert expense_split_asset.is_debit is False
        assert expense_split_inc.is_credit is False
        assert expense_split_inc.is_debit is True



#!/usr/local/bin/python
"""original script from https://github.com/MatzeB/pygnucash/blob/master/gnucash2ledger.py by Matthias Braun matze@braunis.de
 adapted for:
 - python 3 support
 - new string formatting

The security must exist in GnuCash database.
Example of a valid CSV file:
    currency,commodity,value,date
    AUD,YMAX,9.06,"2017-11-10"
 
"""
import argparse
import csv
import sys
from datetime import datetime
from decimal import Decimal

import piecash
from piecash import Price

parser = argparse.ArgumentParser(
    description="""
Import and export prices to and from a gnucash book.

Per default, it exports the full prices list from the gnucash book to the standard output in a CSV format.
To import, add the argument "--import file_to_import.csv".

The format used is a standard CSV (comma as separator) with the following columns:
 - date (YYYY-MM-DD)
 - commodity (gnucash mnemonic)
 - currency (gnucash mnemonic)
 - type (string, optional)
 - value (float)
 - type (string, optional)
""",
    formatter_class=argparse.RawTextHelpFormatter,
)
parser.add_argument("gnucash_filename", help="the name of the gnucash file")
parser.add_argument(
    "--import",
    dest="operation",
    help="to import the prices from a csv file (default export)",
)
args = parser.parse_args()

# args.operation is the name of the import file (.csv).

if args.operation is None:
    # export the prices
    sys.stdout.write(
        "date,type,value,value_num, value_denom, currency,commodity,source\n"
    )
    with piecash.open_book(args.gnucash_filename, open_if_lock=True) as book:
        sys.stdout.writelines(
            "{p.date:%Y-%m-%d},{p.type},{p.value},{p._value_num},{p._value_denom},{p.currency.mnemonic},{p.commodity.mnemonic},{p.source}\n".format(
                p=p
            )
            for p in book.prices
        )
else:
    # import the prices
    with piecash.open_book(
        args.gnucash_filename, open_if_lock=True, readonly=False
    ) as book:
        cdty = book.commodities
        importFile = open(args.operation, "r")

        for l in csv.DictReader(importFile):
            cur = cdty(mnemonic=l["currency"])
            com = cdty(mnemonic=l["commodity"])
            type = l.get("type", None)
            date = datetime.strptime(l["date"], "%Y-%m-%d")
            v = Decimal(l["value"])
            Price(
                currency=cur,
                commodity=com,
                date=date,
                value=v,
                source="piecash-importer",
            )
        book.save()



from __future__ import division

import uuid

from sqlalchemy import Column, VARCHAR, INTEGER, BIGINT, ForeignKey
from sqlalchemy.orm import relation, foreign

from ._common import hybrid_property_gncnumeric, Recurrence, CallableList
from ._declbase import DeclarativeBaseGuid
from .sa_extra import DeclarativeBase


class Budget(DeclarativeBaseGuid):
    """
    A GnuCash Budget

    Attributes:
        name (str): name of the budget
        description (str): description of the budget
        amounts (list of :class:`piecash.budget.BudgetAmount`): list of amounts per account
    """

    __tablename__ = "budgets"

    __table_args__ = {}

    # column definitions
    # keep this line as we reference it in the primaryjoin
    guid = Column(
        "guid",
        VARCHAR(length=32),
        primary_key=True,
        nullable=False,
        default=lambda: uuid.uuid4().hex,
    )
    name = Column("name", VARCHAR(length=2048), nullable=False)
    description = Column("description", VARCHAR(length=2048))
    num_periods = Column("num_periods", INTEGER(), nullable=False)

    # # relation definitions
    recurrence = relation(
        Recurrence,
        primaryjoin=foreign(Recurrence.obj_guid) == guid,
        cascade="all, delete-orphan",
        uselist=False,
    )

    amounts = relation(
        "BudgetAmount",
        back_populates="budget",
        cascade="all, delete-orphan",
        collection_class=CallableList,
    )

    def __str__(self):
        return "Budget<{}({}) for {} periods following pattern '{}' >".format(
            self.name, self.description, self.num_periods, self.recurrence
        )


class BudgetAmount(DeclarativeBase):
    """
    A GnuCash BudgetAmount

    Attributes:
        amount (:class:`decimal.Decimal`): the budgeted amount
        account (:class:`piecash.core.account.Account`): the budgeted account
        budget (:class:`Budget`): the budget of the amount
    """

    __tablename__ = "budget_amounts"

    __table_args__ = {"sqlite_autoincrement": True}

    # column definitions
    id = Column("id", INTEGER(), primary_key=True, autoincrement=True, nullable=False)
    budget_guid = Column(
        "budget_guid", VARCHAR(length=32), ForeignKey("budgets.guid"), nullable=False
    )
    account_guid = Column(
        "account_guid", VARCHAR(length=32), ForeignKey("accounts.guid"), nullable=False
    )
    period_num = Column("period_num", INTEGER(), nullable=False)
    _amount_num = Column("amount_num", BIGINT(), nullable=False)
    _amount_denom = Column("amount_denom", BIGINT(), nullable=False)
    amount = hybrid_property_gncnumeric(_amount_num, _amount_denom)

    # relation definitions
    account = relation("Account", back_populates="budget_amounts")
    budget = relation("Budget", back_populates="amounts")

    def __str__(self):
        return "BudgetAmount<{}={}>".format(self.period_num, self.amount)



#!/usr/bin/env python
# # @file
# @brief Creates a basic set of accounts and a couple of transactions
# @ingroup python_bindings_examples
from decimal import Decimal
import os
import tempfile

from piecash import create_book, Account, Transaction, Split, Commodity
from piecash.core.factories import create_currency_from_ISO

FILE_1 = os.path.join(tempfile.gettempdir(), "example.gnucash")

with create_book(FILE_1, overwrite=True) as book:
    root_acct = book.root_account
    cad = create_currency_from_ISO("CAD")
    expenses_acct = Account(
        parent=root_acct, name="Expenses", type="EXPENSE", commodity=cad
    )
    savings_acct = Account(parent=root_acct, name="Savings", type="BANK", commodity=cad)
    opening_acct = Account(
        parent=root_acct, name="Opening Balance", type="EQUITY", commodity=cad
    )
    num1 = Decimal("4")
    num2 = Decimal("100")
    num3 = Decimal("15")

    # create transaction with core objects in one step
    trans1 = Transaction(
        currency=cad,
        description="Groceries",
        splits=[
            Split(value=num1, account=expenses_acct),
            Split(value=-num1, account=savings_acct),
        ],
    )

    # create transaction with core object in multiple steps
    trans2 = Transaction(currency=cad, description="Opening Savings Balance")

    split3 = Split(value=num2, account=savings_acct, transaction=trans2)

    split4 = Split(value=-num2, account=opening_acct, transaction=trans2)

    # create transaction with factory function
    from piecash.core.factories import single_transaction

    trans3 = single_transaction(
        None, None, "Pharmacy", num3, savings_acct, expenses_acct
    )

    book.save()



import uuid

from sqlalchemy import Column, INTEGER, BIGINT, VARCHAR, ForeignKey
from sqlalchemy.orm import composite, relation

from .person import PersonType

# change of the __doc__ string as getting error in sphinx ==> should be reported to SA project

composite.__doc__ = None  # composite.__doc__.replace(":ref:`mapper_composite`", "")

from ..sa_extra import _DateTime
from .._common import CallableList, hybrid_property_gncnumeric
from .._declbase import DeclarativeBaseGuid


class Billterm(DeclarativeBaseGuid):
    __tablename__ = "billterms"

    __table_args__ = {}

    # column definitions
    guid = Column(
        "guid",
        VARCHAR(length=32),
        primary_key=True,
        nullable=False,
        default=lambda: uuid.uuid4().hex,
    )
    name = Column("name", VARCHAR(length=2048), nullable=False)
    description = Column("description", VARCHAR(length=2048), nullable=False)
    refcount = Column("refcount", INTEGER(), nullable=False)
    invisible = Column("invisible", INTEGER(), nullable=False)
    parent_guid = Column("parent", VARCHAR(length=32), ForeignKey("billterms.guid"))
    type = Column("type", VARCHAR(length=2048), nullable=False)
    duedays = Column("duedays", INTEGER())
    discountdays = Column("discountdays", INTEGER())
    _discount_num = Column("discount_num", BIGINT())
    _discount_denom = Column("discount_denom", BIGINT())
    discount = hybrid_property_gncnumeric(_discount_num, _discount_denom)
    cutoff = Column("cutoff", INTEGER())

    # relation definitions
    children = relation(
        "Billterm",
        back_populates="parent",
        cascade="all, delete-orphan",
        collection_class=CallableList,
    )
    parent = relation(
        "Billterm",
        back_populates="children",
        remote_side=guid,
    )


class Entry(DeclarativeBaseGuid):
    __tablename__ = "entries"

    __table_args__ = {}

    # column definitions
    date = Column("date", _DateTime(), nullable=False)
    date_entered = Column("date_entered", _DateTime())
    description = Column("description", VARCHAR(length=2048))
    action = Column("action", VARCHAR(length=2048))
    notes = Column("notes", VARCHAR(length=2048))
    _quantity_num = Column("quantity_num", BIGINT())
    _quantity_denom = Column("quantity_denom", BIGINT())
    _quantity_denom_basis = None
    quantity = hybrid_property_gncnumeric(_quantity_num, _quantity_denom)

    i_acct = Column("i_acct", VARCHAR(length=32))
    _i_price_num = Column("i_price_num", BIGINT())
    _i_price_denom = Column("i_price_denom", BIGINT())
    _i_price_denom_basis = None
    i_price = hybrid_property_gncnumeric(_i_price_num, _i_price_denom)

    _i_discount_num = Column("i_discount_num", BIGINT())
    _i_discount_denom = Column("i_discount_denom", BIGINT())
    _i_discount_denom_basis = None
    i_discount = hybrid_property_gncnumeric(_i_discount_num, _i_discount_denom)

    invoice_guid = Column("invoice", VARCHAR(length=32), ForeignKey("invoices.guid"))
    i_disc_type = Column("i_disc_type", VARCHAR(length=2048))
    i_disc_how = Column("i_disc_how", VARCHAR(length=2048))
    i_taxable = Column("i_taxable", INTEGER())
    i_taxincluded = Column("i_taxincluded", INTEGER())
    i_taxtable = Column("i_taxtable", VARCHAR(length=32))

    b_acct = Column("b_acct", VARCHAR(length=32))
    _b_price_num = Column("b_price_num", BIGINT())
    _b_price_denom = Column("b_price_denom", BIGINT())
    _b_price_denom_basis = None
    b_price = hybrid_property_gncnumeric(_b_price_num, _b_price_denom)
    bill = Column("bill", VARCHAR(length=32))
    b_taxable = Column("b_taxable", INTEGER())
    b_taxincluded = Column("b_taxincluded", INTEGER())
    b_taxtable = Column("b_taxtable", VARCHAR(length=32))
    b_paytype = Column("b_paytype", INTEGER())
    billable = Column("billable", INTEGER())
    billto_type = Column("billto_type", INTEGER())
    billto_guid = Column("billto_guid", VARCHAR(length=32))
    order_guid = Column("order_guid", VARCHAR(length=32), ForeignKey("orders.guid"))

    # relation definitions
    order = relation("Order", back_populates="entries")
    invoice = relation("Invoice", back_populates="entries")

    def __str__(self):
        return "Entry<{}>".format(self.description)


class Invoice(DeclarativeBaseGuid):
    __tablename__ = "invoices"

    __table_args__ = {}

    # column definitions
    id = Column("id", VARCHAR(length=2048), nullable=False)
    date_opened = Column("date_opened", _DateTime())
    date_posted = Column("date_posted", _DateTime())
    notes = Column("notes", VARCHAR(length=2048), nullable=False)
    active = Column("active", INTEGER(), nullable=False)
    currency_guid = Column(
        "currency", VARCHAR(length=32), ForeignKey("commodities.guid"), nullable=False
    )
    owner_type = Column("owner_type", INTEGER())
    owner_guid = Column("owner_guid", VARCHAR(length=32))

    # owner = generic_relationship(owner_type, owner_guid,
    #                              map_type2discriminator={"Customer": 2,
    #                                                      "Vendor": 1,
    #                                                      })
    term_guid = Column("terms", VARCHAR(length=32), ForeignKey("billterms.guid"))
    billing_id = Column("billing_id", VARCHAR(length=2048))
    post_txn_guid = Column("post_txn", VARCHAR(length=32), ForeignKey("transactions.guid"))
    post_lot_guid = Column(
        "post_lot", VARCHAR(length=32), ForeignKey("lots.guid")
    )
    post_acc_guid = Column("post_acc", VARCHAR(length=32), ForeignKey("accounts.guid"))
    billto_type = Column("billto_type", INTEGER())
    billto_guid = Column("billto_guid", VARCHAR(length=32))
    _charge_amt_num = Column("charge_amt_num", BIGINT())
    _charge_amt_denom = Column("charge_amt_denom", BIGINT())
    charge_amt = hybrid_property_gncnumeric(_charge_amt_num, _charge_amt_denom)

    # relation definitions
    # todo: check all relations and understanding of types...
    term = relation("Billterm")
    currency = relation("Commodity")
    post_account = relation("Account")
    post_lot = relation("Lot")
    post_txn = relation("Transaction")

    entries = relation(
        "Entry",
        back_populates="invoice",
        cascade="all, delete-orphan",
        collection_class=CallableList,
    )

    def __str__(self):
        return "Invoice<{}>".format(self.id)


class Job(DeclarativeBaseGuid):
    __tablename__ = "jobs"

    __table_args__ = {}

    # column definitions
    id = Column("id", VARCHAR(length=2048), nullable=False)
    name = Column("name", VARCHAR(length=2048), nullable=False)
    reference = Column("reference", VARCHAR(length=2048), nullable=False)
    active = Column("active", INTEGER(), nullable=False)
    owner_type = Column("owner_type", INTEGER())
    owner_guid = Column("owner_guid", VARCHAR(length=32))

    # relation definitions
    # todo: owner_guid/type links to Vendor or Customer
    def __init__(self, name, owner, reference="", active=1):
        self.name = name
        self.reference = reference
        self.active = active
        self.owner_type = PersonType[type(owner)]
        self.owner_guid = owner.guid
        if owner.book:
            owner.book.add(self)

    def __str__(self):
        return "Job<{self.name}>".format(self=self)

    def on_book_add(self):
        self._assign_id()

    # hold the name of the counter to use for id
    _counter_name = "counter_job"

    def _assign_id(self):
        if not self.id:
            cnt = getattr(self.book, self._counter_name) + 1
            setattr(self.book, self._counter_name, cnt)
            self.id = "{:06d}".format(cnt)


# This class exists in code but not in the GUI (to confirm?)


class Order(DeclarativeBaseGuid):
    __tablename__ = "orders"

    __table_args__ = {}

    # column definitions
    id = Column("id", VARCHAR(length=2048), nullable=False)
    notes = Column("notes", VARCHAR(length=2048), nullable=False)
    reference = Column("reference", VARCHAR(length=2048), nullable=False)
    active = Column("active", INTEGER(), nullable=False)

    date_opened = Column("date_opened", _DateTime(), nullable=False)
    date_closed = Column("date_closed", _DateTime(), nullable=False)
    owner_type = Column("owner_type", INTEGER(), nullable=False)
    owner_guid = Column("owner_guid", VARCHAR(length=32), nullable=False)

    # relation definitions
    # todo: owner_guid/type links to Vendor or Customer
    entries = relation(
        "Entry",
        back_populates="order",
        cascade="all, delete-orphan",
        collection_class=CallableList,
    )



from __future__ import unicode_literals

import json
import re
from functools import singledispatch
from locale import getdefaultlocale

from .core import Account, Book, Commodity, Price, Transaction

"""original script from https://github.com/MatzeB/pygnucash/blob/master/gnucash2ledger.py by Matthias Braun matze@braunis.de
 adapted for:
 - python 3 support
 - new string formatting
"""

try:
    import babel
    import babel.numbers

    BABEL_AVAILABLE = True
except ImportError:
    BABEL_AVAILABLE = False


@singledispatch
def ledger(obj, **kwargs):
    raise NotImplemented


CURRENCY_RE = re.compile("^[A-Z]{3}$")
NUMBER_RE = re.compile(r"(^|\s+)[-+]?([0-9]*(\.|,)[0-9]+|[0-9]+)($|\s+)")  # regexp to identify a float with blank spaces
NUMERIC_SPACE = re.compile(r"[0-9\s]")
CHARS_ONLY = re.compile(r"[A-Za-z]+")


def quote_commodity(mnemonic):
    if not CHARS_ONLY.fullmatch(mnemonic):
        return json.dumps(mnemonic)
    else:
        return mnemonic


def format_commodity(mnemonic, locale):
    if CURRENCY_RE.match(mnemonic) or True:
        # format the currency via BABEL if available and then remove the number/amount
        s = format_currency(0, 0, mnemonic, locale)
        return NUMBER_RE.sub("", s)
    else:
        # just quote the commodity
        return quote_commodity(mnemonic)


def format_currency(amount, decimals, currency, locale=False, decimal_quantization=True):
    currency = quote_commodity(currency)
    if locale is True:
        locale = getdefaultlocale()[0]
        if BABEL_AVAILABLE is False:
            raise ValueError(f"You must install babel ('pip install babel') to export to ledger in your locale '{locale}'")
        else:
            return babel.numbers.format_currency(
                amount,
                currency,
                format=None,
                locale=locale,
                currency_digits=True,
                format_type="standard",
                decimal_quantization=decimal_quantization,
            )
    else:
        # local hand made version
        if decimal_quantization:
            digits = decimals
        else:
            digits = max(decimals, len((str(amount) + ".").split(".")[1].rstrip("0")))
        return "{} {:,.{}f}".format(currency, amount, digits)


@ledger.register(Transaction)
def _(tr, locale=False, **kwargs):
    """Return a ledger-cli alike representation of the transaction"""
    s = [
        "{:%Y-%m-%d} {}{}\n".format(
            tr.post_date,
            "({}) ".format(tr.num.replace(")", "")) if tr.num else "",
            tr.description,
        )
    ]
    if tr.notes:
        s.append("\t;{}\n".format(tr.notes))
    for split in sorted(
        tr.splits,
        key=lambda split: (split.value, split.transaction_guid, split.account_guid),
    ):
        if split.account.commodity.mnemonic == "template":
            return ""
        if split.reconcile_state in ["c", "y"]:
            s.append("\t* {:38}  ".format(split.account.fullname))
        else:
            s.append("\t{:40}  ".format(split.account.fullname))
        if split.account.commodity != tr.currency:
            s.append(
                "{quantity} @@ {amount}".format(
                    quantity=format_currency(
                        split.quantity,
                        split.account.commodity.precision,
                        split.account.commodity.mnemonic,
                        locale,
                        decimal_quantization=False,
                    ),
                    amount=format_currency(
                        abs(split.value),
                        tr.currency.precision,
                        tr.currency.mnemonic,
                        locale,
                    ),
                )
            )
        else:
            s.append(format_currency(split.value, tr.currency.precision, tr.currency.mnemonic, locale))

        if split.memo:
            s.append(" ;   {:20}".format(split.memo))
        s.append("\n")

    return "".join(s)


@ledger.register(Commodity)
def _(cdty, locale=False, commodity_notes=False, **kwargs):
    """Return a ledger-cli alike representation of the commodity"""
    if cdty.mnemonic in ["", "template"]:
        return ""
    res = "commodity {}\n".format(format_commodity(cdty.mnemonic, locale))
    if cdty.fullname != "" and commodity_notes:
        res += "\tnote {}\n".format(cdty.fullname, locale)
    res += "\n"
    return res


@ledger.register(Account)
def _(acc, short_account_names=False, locale=False, **kwargs):
    """Return a ledger-cli alike representation of the account"""
    # ignore "dummy" accounts
    if acc.type is None or acc.parent is None:
        return ""
    if acc.commodity.mnemonic == "template":
        return ""

    if short_account_names:
        res = "account {}\n".format(acc.name)
    else:
        res = "account {}\n".format(acc.fullname)

    if acc.description != "":
        res += "\tnote {}\n".format(acc.description)

    if acc.commodity.is_currency():
        res += '\tcheck commodity == "{}"\n'.format(format_commodity(acc.commodity.mnemonic,locale))  # .replace('"', '\\"'))
    return res


@ledger.register(Price)
def _(price, locale=False, **kwargs):
    """Return a ledger-cli alike representation of the price"""
    return "P {:%Y-%m-%d %H:%M:%S} {} {}\n".format(
        price.date,
        format_commodity(price.commodity.mnemonic, locale),
        format_currency(
            price.value,
            price.currency.precision,
            price.currency.mnemonic,
            locale,
            decimal_quantization=False,
        ),
    )


@ledger.register(Book)
def _(book, **kwargs):
    """Return a ledger-cli alike representation of the book"""
    res = []

    # Commodities
    for commodity in sorted(book.commodities, key=lambda cdty: cdty.mnemonic):
        res.append(ledger(commodity, **kwargs))

    # Accounts
    if kwargs.get("short_account_names"):  # check that no ambiguity in account names
        accounts = [acc.name for acc in book.accounts]
        if len(accounts) != len(set(accounts)):
            raise ValueError("You have duplicate short names in your book. " "You cannot use the 'short_account_names' option.")
    for acc in book.accounts:
        res.append(ledger(acc, **kwargs))
        res.append("\n")

    # Prices
    for price in sorted(book.prices, key=lambda x: (x.commodity_guid, x.currency_guid, x.date)):
        res.append(ledger(price, **kwargs))
    res.append("\n")

    for trans in sorted(book.transactions, key=lambda x: x.post_date):
        res.append(ledger(trans, **kwargs))
        res.append("\n")

    return "".join(res)



from decimal import Decimal

from piecash import ledger
from test_helper import (
    db_sqlite_uri,
    db_sqlite,
    new_book,
    new_book_USD,
    book_uri,
    book_transactions,
    book_complex,
)

# dummy line to avoid removing unused symbols
a = (
    db_sqlite_uri,
    db_sqlite,
    new_book,
    new_book_USD,
    book_uri,
    book_transactions,
    book_complex,
)


def test_get_balance(book_complex):
    """
    Tests listing the commodity quantity in the account.
    """

    asset = book_complex.accounts.get(name="Asset")
    broker = book_complex.accounts.get(name="Broker")
    foo_stock = book_complex.accounts.get(name="Foo stock")
    expense = book_complex.accounts.get(name="Expense")
    income = book_complex.accounts.get(name="Income")
    assert foo_stock.get_balance(recurse=True) == Decimal("130")
    assert broker.get_balance(recurse=True) == Decimal("117")
    assert asset.get_balance(recurse=False) == Decimal("0")
    assert asset.get_balance() == Decimal("24695.3")
    assert expense.get_balance() == Decimal("260")
    assert income.get_balance() == Decimal("150")
    assert income.get_balance(natural_sign=False) == Decimal("-150")
